//
//  DocumentViewer.m
//

#import "DocumentViewer.h"
#import "FileURLHelper.h"
#import <CommonCrypto/CommonDigest.h>

@implementation NSString (SHA)

- (NSString *)sha1
{
	NSData *data = [self dataUsingEncoding:NSUTF8StringEncoding];
	uint8_t digest[CC_SHA1_DIGEST_LENGTH];
	
	CC_SHA1(data.bytes, (CC_LONG)(data.length), digest);
	
	NSMutableString *output = [NSMutableString stringWithCapacity:CC_SHA1_DIGEST_LENGTH * 2];
	
	for (int i = 0; i < CC_SHA1_DIGEST_LENGTH; i++)
	{
		[output appendFormat:@"%02x", digest[i]];
	}
	
	return output;
}
@end

@implementation DocumentViewer {
	NSArray * _documentURLStrings;
	NSMutableDictionary * _requestsByURLString;
	NSMutableDictionary * _pathsByIndex;
}

- (void)awakeFromNib {
	[super awakeFromNib];
	// try again if the previous initialization failed due to an unset viewcontroller:
	[self setDocumentURLs:_documentURLStrings];
}

- (void)dealloc {
	_previewController = nil;
}

- (NSInteger)numberOfPreviewItemsInPreviewController:(QLPreviewController *)controller {
	//NSLog(@"_documentURLStrings.count: %d", _documentURLStrings.count);
	return _documentURLStrings.count;
	//return 1;
}


- (id <QLPreviewItem>)previewController:(QLPreviewController *)controller previewItemAtIndex:(NSInteger)index {

	NSInteger c = _documentURLStrings.count;
	if (index < 0 || index > c-1) {
		return nil;
	}
	//index = 0;
	
	NSString * urlString = _documentURLStrings[index];
	
	if ([urlString hasPrefix:@"file:"]) {
		//NSString * filename = [urlString substringFromIndex:5];
		NSString * path = GetFileURLFromString(urlString).path;
		//[[NSBundle mainBundle] pathForResource:filename ofType:nil];
		if (path) return [NSURL fileURLWithPath:path];
	}
	else if ([urlString hasPrefix:@"http:"] || [urlString hasPrefix:@"https:"]) {
		
		if (!_pathsByIndex) _pathsByIndex = @{}.mutableCopy;
		NSString * cacheFilePath = _pathsByIndex[[NSNumber numberWithInteger:index]];
		NSString *dummyFilePath = nil;
		if (!cacheFilePath) {
			NSArray * cachePathList = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
			NSString * cachePath    = [cachePathList objectAtIndex:0];
			NSString * cacheFilePath = [cachePath stringByAppendingPathComponent:[urlString sha1]];
			if (urlString.pathExtension.length) {
				cacheFilePath = [cacheFilePath stringByAppendingPathExtension:urlString.pathExtension];
			}
			_pathsByIndex[[NSNumber numberWithInteger:index]] = cacheFilePath;
			dummyFilePath = [cachePath stringByAppendingPathComponent:@"placeholder.jpg"];
		}
		
		[self createDummyDocumentAtPath:dummyFilePath];
		
		if ([[NSFileManager defaultManager] fileExistsAtPath:cacheFilePath isDirectory:nil]) {
			return [NSURL fileURLWithPath:cacheFilePath];
		}
		
		if (!_requestsByURLString) _requestsByURLString = @{}.mutableCopy;
		if (_requestsByURLString[urlString]) return nil;
		
		//NSData * data = [NSData dataWithContentsOfURL:[NSURL URLWithString:urlString]];
		NSURLRequest * request = [NSURLRequest
								  requestWithURL:[NSURL URLWithString:urlString]
								  cachePolicy:NSURLRequestReturnCacheDataElseLoad
								  timeoutInterval:5.0];
		_requestsByURLString[urlString] = request;
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
		
		[NSURLConnection sendAsynchronousRequest:request queue:[NSOperationQueue mainQueue] completionHandler:^(NSURLResponse *response, NSData *data, NSError *connectionError) {
			
			if (connectionError || !data) return;
			if (!self.previewController) return;
			
			[data writeToFile:cacheFilePath atomically:NO];
			[_requestsByURLString removeObjectForKey:urlString];
			if([self previewController:self.previewController previewItemAtIndex:index] != nil) {
				[self.previewController reloadData];
			}
			
		}];
#pragma GCC diagnostic pop
		if(dummyFilePath) return [NSURL fileURLWithPath:dummyFilePath];
	}
	return nil;

}

- (void)createDummyDocumentAtPath:(NSString*)path {
	if (!path || [[NSFileManager defaultManager] isReadableFileAtPath:path]) {
		return;
	}
	CGSize size = CGSizeMake(1, 1);
	UIGraphicsBeginImageContextWithOptions(size, NO, 0.0);
	//[[UIColor redColor] setFill];
	//UIRectFill(CGRectMake(0, 0, size.width, size.height));
	UIImage *image = UIGraphicsGetImageFromCurrentImageContext();
	UIGraphicsEndImageContext();
	[UIImagePNGRepresentation(image) writeToFile:path atomically:YES];
}

- (void)setDocumentURLs:(id)values {
	if ([values isKindOfClass:[NSString class]]) {
		_documentURLStrings = [((NSString *)values) componentsSeparatedByString:@"\n"];
	} else if ([values isKindOfClass:[NSArray class]] && [values count] > 0) {
		_documentURLStrings = values;
	} else {
		return;
	}
	if(!self.viewController) {
		return;
	}
	if (!self.previewController.parentViewController) {
		QLPreviewController * preview = [[QLPreviewController alloc] init];
		self.previewController = preview;
		[self.viewController addChildViewController:self.previewController];
		//set the frame from the parent view
		CGFloat w = self.frame.size.width;
		CGFloat h = self.frame.size.height;
		self.previewController.view.frame = CGRectMake(0, 0, w, h);
		self.previewController.view.autoresizingMask = UIViewAutoresizingFlexibleWidth|UIViewAutoresizingFlexibleHeight;
		if (!self.previewController.view.superview) {
			[self.viewController.view addSubview:self.previewController.view];
			[self.previewController didMoveToParentViewController:self.viewController];
		}
		//save a reference to the preview controller in an ivar
	}
	self.previewController.dataSource = self;
	self.previewController.delegate = self;
	[self.previewController reloadData];
	
}

@end
