//
//  PhotoZoom.m
//
//

#import "PhotoZoom.h"
#import "PZViewController.h"
#import "PZPagingScrollView.h"
#import "PZPhotosDataSource.h"
#import "FileURLHelper.h"

@protocol InputDataSource <NSObject>
- (NSArray *)collectedValuesForKeyPath:(NSString *)keyPath;
- (void)navigateToItemAtIndex:(NSUInteger)index;
@end

@protocol ViewControllerAdditions <NSObject>
@property BOOL prefersStatusBarHidden;
@end

@protocol DataSourceBase <NSObject>
- (NSInteger)numberOfItemsInSection:(NSInteger)section;
- (NSDictionary *)datasetForItemAtIndexPath:(NSIndexPath *)indexPath;
@end


@implementation PhotoZoom {
	PZPhotosDataSource * _photosDataSource;
	CGSize _lastSize;
	NSArray * _imageInfos;
	int _spaceBetweenImages;
	PZPagingScrollView * _scrollView;
	BOOL _isFullScreen;
	BOOL _isNavigating;
	BOOL _isInNavigationMode;
	NSArray * _bindingsInfo;
}

- (void)setImageURLs:(id)value {
	if (_imageInfos && _isInNavigationMode) {
		// don't change the values in navigation mode.
		// instead, scroll to the according image.
		if (_isNavigating) return;
		
		[self setSelectedImageURL:value];
		
		return;
	}
	NSArray * urlStrings;
	if ([value isKindOfClass:[NSString class]]) {
		urlStrings = [[((NSString *)value) stringByTrimmingCharactersInSet:NSCharacterSet.whitespaceAndNewlineCharacterSet] componentsSeparatedByString:@"\n"];
	}
	else if ([value isKindOfClass:[NSArray class]]) {
		NSMutableArray * array = @[].mutableCopy;
		for (id url in value) {
			if ([url isKindOfClass:[NSString class]])
				[array addObject:url];
		}
		urlStrings = array;
	}
	else if ([value isKindOfClass:NSClassFromString(@"DataSourceBase")]) {
		id<DataSourceBase> ds = (id<DataSourceBase>)value;
		NSMutableArray * array = @[].mutableCopy;
		for (NSInteger i = 0, c = [ds numberOfItemsInSection:0]; i < c; i++) {
			NSString * s = [NSString stringWithFormat:@"%@", [ds datasetForItemAtIndexPath:[NSIndexPath indexPathForRow:i inSection:0]]];
			[array addObject:s];
		}
		urlStrings = array;
	}
	else return;
	
	NSMutableArray * imageInfos = @[].mutableCopy;
	for (NSString * urlString in urlStrings) {
		if (![urlString isKindOfClass:[NSString class]])
			continue;
		[imageInfos addObject:@{@"URL":urlString}.mutableCopy];
	}
	_imageInfos = imageInfos;
	if (_scrollView) {
		[_scrollView removeFromSuperview];
		[self createScrollView];
	}
}

- (void)setSpaceBetweenImages:(int)space {
	_spaceBetweenImages = space;
}

- (void)setSelectedImageURL:(NSString *)selectedImageURL {
	if (_isNavigating) return;
	int index = 0;
	for (NSDictionary * info in _imageInfos) {
		if ([info[@"URL"] isEqualToString:selectedImageURL]) break;
		index++;
	}
	if (index == _imageInfos.count) return;
	[_scrollView displayPagingViewAtIndex:index];
}

- (NSString *)selectedImageURL {
	NSDictionary * info = _imageInfos[_scrollView.currentPagingIndex];
	return info ? info[@"URL"] : nil;
}

- (void)awakeFromNib {
	[super awakeFromNib];
	self.clipsToBounds = YES;
    _photosDataSource = [[PZPhotosDataSource alloc] init];
	_lastSize = self.frame.size;
	
	[self prepareForNavigation];
	[self createScrollView];
}

- (void)prepareForNavigation {
	if (!_bindingsInfo || !_inputDataSource) return;
	// find the 'imageURLs' binding
	for (NSDictionary * bindingInfo in _bindingsInfo) {
		if ([bindingInfo[@"sourceKeyPath"] isEqualToString:@"imageURLs"]) {
			if ([bindingInfo[@"destination"] isEqualToString:[_inputDataSource valueForKey:@"identifier"]]) {
				NSArray * values = [_inputDataSource performSelector:@selector(collectedValuesForKeyPath:) withObject:bindingInfo[@"destinationKeyPath"]];
				[self setImageURLs:values];
				_isInNavigationMode = YES;
			}
		}
	}
}

- (void)createScrollView {
	CGRect frame = self.bounds;
	frame.size.width += _spaceBetweenImages;
	_scrollView = [PZPagingScrollView.alloc initWithFrame:frame];
	_scrollView.spaceBetweenImages = _spaceBetweenImages;
	[self addSubview:_scrollView];
	_scrollView.pagingViewDelegate = self;
	if (_imageInfos.count) {
		[_scrollView resetDisplay];
		[_scrollView tilePages];
	}
}


- (void)layoutSubviews {
	if (CGSizeEqualToSize(self.frame.size, _lastSize))
		return;
	_lastSize = self.frame.size;
	
	_scrollView.suspendTiling = FALSE;
    [_scrollView resetDisplay];
}

- (void)photoForIndex:(NSUInteger)index withCompletionBlock:(void (^)(NSDictionary *, NSError *))completionBlock {
	NSMutableDictionary * info = _imageInfos[index];
	if (info[@"image"]) {
		completionBlock(info, nil);
		return;
	}
	NSString * urlString = [_imageInfos[index] objectForKey:@"URL"];
	
	if ([urlString hasPrefix:@"file:"]) {
		info[@"image"] = [UIImage imageWithContentsOfFile:GetFileURLFromString(urlString).path];
		//[UIImage imageNamed:[urlString substringFromIndex:5]];
		completionBlock(info, nil);
		return;
	}
	
	[[[NSURLSession sharedSession] dataTaskWithURL:[NSURL URLWithString:urlString]
								 completionHandler:^(NSData *data,
													 NSURLResponse *response,
													 NSError *error) {
									 if (error || !data.length) {
										 completionBlock(nil, error);
										 return;
									 }
									 else {
										 info[@"image"] = [UIImage imageWithData:data];
										 completionBlock(info, nil);
									 }
								 }] resume];
}

- (BOOL)photoIsReadyForIndex:(NSInteger)index {
	NSMutableDictionary * info = _imageInfos[index];
	return info[@"image"] != nil;
}

- (void)toggleNavigationBarVisibility {
	
	if (!_viewController.navigationController.navigationBar.translucent) return;
	
    if (_isFullScreen) {
        // fade in navigation
        
		if ([_viewController respondsToSelector:@selector(setPrefersStatusBarHidden:)]) {
			[_viewController setValue:@NO forKey:@"prefersStatusBarHidden"];
		}
		_viewController.navigationController.navigationBarHidden = NO;
		_viewController.navigationController.navigationBar.alpha = 0;
		
        [UIView animateWithDuration:0.2 animations:^{
			if ([_viewController respondsToSelector:@selector(setNeedsStatusBarAppearanceUpdate)])
				[_viewController setNeedsStatusBarAppearanceUpdate];
           // [[UIApplication sharedApplication] setStatusBarHidden:FALSE withAnimation:UIStatusBarAnimationNone];
            _viewController.navigationController.navigationBar.alpha = 1.0;
        } completion:^(BOOL finished) {
        }];
    }
    else {
        // fade out navigation
		
        [UIView animateWithDuration:0.2 animations:^{
          //  [[UIApplication sharedApplication] setStatusBarHidden:TRUE withAnimation:UIStatusBarAnimationFade];
            _viewController.navigationController.navigationBar.alpha = 0.0;
        } completion:^(BOOL finished) {
			if (!finished) return;
			_viewController.navigationController.navigationBarHidden = YES;
			if ([_viewController respondsToSelector:@selector(setPrefersStatusBarHidden:)]) {
				[_viewController setValue:@YES forKey:@"prefersStatusBarHidden"];
				if ([_viewController respondsToSelector:@selector(setNeedsStatusBarAppearanceUpdate)])
					[_viewController setNeedsStatusBarAppearanceUpdate];
			}
        }];
	}
	
}

- (void)toggleTabBarVisibility {
	
	if (![_viewController.tabBarController.tabBar respondsToSelector:@selector(isTranslucent)]) return;
	if (!_viewController.tabBarController.tabBar.translucent) return;
	
	if (_isFullScreen) {
        // fade in tab bar
		
		[UIView animateWithDuration:0.2 animations:^{
            _viewController.tabBarController.tabBar.alpha = 1;
        } completion:^(BOOL finished) {}];
		
	} else {
        // fade out tab bar
		
		[UIView animateWithDuration:0.2 animations:^{
            _viewController.tabBarController.tabBar.alpha = 0;
        } completion:^(BOOL finished) {}];
	}
	
}

- (void)toggleFullScreen {
	if (self.bounds.origin.y > 0)
		return;
	
	[self toggleNavigationBarVisibility];
	[self toggleTabBarVisibility];
	
	_isFullScreen = !_isFullScreen;
}

#pragma mark - Navigation

- (void)setInputDataSource:(id)inputDataSource {
	_inputDataSource = inputDataSource;
	
}

- (void)setBindingsInfoString:(NSString *)bindingsInfoString {
	_bindingsInfo = [NSJSONSerialization JSONObjectWithData:[bindingsInfoString dataUsingEncoding:NSUTF8StringEncoding] options:0 error:nil];
}


#pragma mark - PZPagingScrollViewDelegate

- (Class)pagingScrollView:(PZPagingScrollView *)pagingScrollView classForIndex:(NSUInteger)index {
    return [PZPhotoView class];
}

- (NSUInteger)pagingScrollViewPagingViewCount:(PZPagingScrollView *)pagingScrollView {
	return _imageInfos.count;
}

- (UIView *)pagingScrollView:(PZPagingScrollView *)pagingScrollView pageViewForIndex:(NSUInteger)index {
    PZPhotoView *photoView = [[PZPhotoView alloc] initWithFrame:self.bounds];
    photoView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
    photoView.photoViewDelegate = self;
	
	//NSLog(@"%@, %@", NSStringFromCGSize(self.contentSize), NSStringFromCGSize(self.bounds.size));
    
    return photoView;
}

- (void)pagingScrollView:(PZPagingScrollView *)pagingScrollView preparePageViewForDisplay:(UIView *)pageView forIndex:(NSUInteger)index {
	
	NSAssert([pageView isKindOfClass:[PZPhotoView class]], @"Invalid State");
	NSAssert(index < _imageInfos.count, @"Invalid State");
	
	//self.title = [_photosDataSource titleForIndex:index];
	BOOL isInstantLoad = [self photoIsReadyForIndex:index];
	
	PZPhotoView *photoView = (PZPhotoView *)pageView;
	
	if (!isInstantLoad) {
		[photoView startWaiting];
		//photoView.activityIndicator.center = CGRectApplyAffineTransform(self.bounds, CGAffineTransformMakeScale(0, 0)).size;
	}
	
	[self photoForIndex:index withCompletionBlock:^(NSDictionary * photoInfo, NSError *error) {
		if (!isInstantLoad) [photoView stopWaiting];
		if (error != nil) {
			// DebugLog(@"Error: %@", error);
		}
		else {
			[photoView displayImage:photoInfo[@"image"] animated:!isInstantLoad];
			//self.title = photoInfo[@"name"];
		}
	}];
	
}

- (void)pagingScrollViewDidChangeCurrentPagingIndex:(PZPagingScrollView *)pagingScrollView {

	SEL sel = @selector(navigateToItemAtIndex:);
	
	if (_inputDataSource && [_inputDataSource respondsToSelector:sel]) {
		_isNavigating = YES;
		
		NSUInteger index = pagingScrollView.currentPagingIndex;
		NSInvocation *inv = [NSInvocation invocationWithMethodSignature:[_inputDataSource methodSignatureForSelector:sel]];
		[inv setSelector:sel];
		[inv setTarget:_inputDataSource];
		[inv setArgument:&index atIndex:2]; //arguments 0 and 1 are self and _cmd respectively, automatically set by NSInvocation
		[inv invoke];
		
		_isNavigating = NO;
	}
}


#pragma mark - PZPhotoViewDelegate

- (void)photoViewDidSingleTap:(PZPhotoView *)photoView {
    [self toggleFullScreen];
}

- (void)photoViewDidDoubleTap:(PZPhotoView *)photoView {
	
}

- (void)photoViewDidTwoFingerTap:(PZPhotoView *)photoView {
	
}

- (void)photoViewDidDoubleTwoFingerTap:(PZPhotoView *)photoView {
	
}

#pragma mark - Context menu support

- (BOOL)canBecomeFirstResponder {
	return YES;
}

- (void)save:(id)sender {
	NSDictionary * info = _imageInfos[_scrollView.currentPagingIndex];
	if (!info) return;
	UIImage * image = info[@"image"];
	if (!image) return;
	UIImageWriteToSavedPhotosAlbum(image, nil, nil, NULL);
}

@end
